% This script implements the whitened pCN algorithm for sampling the posterior
% arising from the OT/HT(alpha) priors based on the Cauchy distribution, in
% Nonparametric Binary Classification with a homogeneously smooth truth. 
% It reproduces the results in Figures D.12 and D.13 in the supplement of Agapiou and Castillo 2024+ 
% "Heavy-tailed Bayesian Nonparametric Adaptation" arXiv:2308.04916

rng(3)
clear all
close all
format long

% Code requires Wavelab850 installed
% Check that computer has Wavelab850 running
run('/Applications/MATLAB_R2022b.app/toolbox/Wavelab850/WavePath.m');

%%%% Data %%%%
%%%%%%%%%%%%%%%%%%%%%%%%
K = 2^10; % finest wavelet level 
K1 = 0:K-1;
K1(1) = 1;
L = 1; % coarsest wavelet level
x=[1/K:1/K:1];


%%%% Wavelet transform %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
QMF8 = MakeONFilter('Daubechies',8); %generate ON wavelet filter for computing wavelet transform
build = @(U) sqrt(K)*IWT_PO(U,L,QMF8); %perform inverse wavelet transform

%%%% Truth %%%%
%%%%%%%%%%%%%%%
wc_truth=(4*(cos(1:K).^3)).*2.^(-(5/2)*floor(log2(K1))); % define true binary regression function
% f0(x)=P(Y=1|X=x) via the coefficients of its logit, wrt to Daubechies 8 wavelet basis
logit_truth=build(wc_truth);
truth=(1./(1+exp(-logit_truth)))'; %Sobolev regularity beta=1
figure(1)
plot(x,truth, 'LineWidth', 1.75)

%%%% Generate Data %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
n=10^4; %number of observations

% For X_i uniformly drawn in [0,1], we generate binary Y_i such that
% P(Y_i=1|X_i)=f0(x)
data=NaN(2,n);
for i=1:n
    x_ind=ceil(K*rand);
    if rand<truth(x_ind)
        data(2,i)=1;
    else
        data(2,i)=0;
    end
    data(1,i)=x_ind;
end
fig2=figure(2)
fig2.Position(3:4)=[500,400];
box on
hold on
plot(data(1,:)/K, data(2,:),'|','Linewidth', 2)
plot(x,truth,'LineWidth', 1.75, 'Color', 'black', 'LineStyle','--');
hold off



%%%% Prior scaling %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
jj = 2.^floor(log2(K1));
% HT(alpha)
%alpha=5;
%scale_vec = ((jj.^(-1/2-alpha)));%prior scaling s_l

% OT
scale_vec=2.^(-(log2(jj)).^(2)); %prior scaling s_l

%%%% Rho or regular posterior %%%%
rho=1; %rho posterior, rho=1 means usual posterior


% Mapping from Gaussian to Cauchy (inverse CDF), needed for whitened
% version of pCN, see Chen et al https://arxiv.org/abs/1803.03344
T = @(xi) 1./(1+exp(-build(scale_vec.*tan((pi/2)*(1-2*normcdf(xi)))))); %Cauchy
%white noise representation
% For Gaussian prior use
%T = @(xi) 1./(1+exp(-build((scale_vec.*xi)))); %Gaussian

%%%% wpCN MCMC parameters %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%
I=200000; %number of iterations
Ir=I*0.875; %number of iterations discarded as burn in

step_size=0.02; %step size tuned to give around 30% acceptance probability
step_compl=sqrt(1-step_size^2);

post_draws_cauchy=NaN(I-Ir,K);

%%%% Initialization %%%%
%%%%%%%%%%%%%%%%%%%%%%%%
old=randn(1,K); %white noise
t_old=T(old); %initialize from the prior
h_old=t_old(data(1,:)); %this is the corresponding binary regression function
%evaluated at the locations x where we have observations
post_draws_cauchy(1,:)=t_old;

acc=0; %acceptance counter

%%%% MCMC %%%%
%%%%%%%%%%%%%%
for i=2:I
  eta=randn(1,K);
  proposal=step_compl*old+step_size*eta;
  t_proposal=T(proposal);
  h_proposal=t_proposal(data(1,:));
  
  laccept_prob=rho*(log(h_proposal)*data(2,:)'+log(1-h_proposal)*(1-data(2,:)')-log(h_old)*data(2,:)'-log(1-h_old)*(1-data(2,:)')); %log of acceptance probability for wpCN
  %see §3.3 of the article for the likelihood in binary regression

  if log(rand) <= laccept_prob
    old=proposal;
    t_old=t_proposal;
    h_old=h_proposal;
    acc=acc+1;
  end
  if(i>Ir) % only store samples after burn-in
    post_draws_cauchy(i-Ir,:)=t_old;
  end
  if round(i/1000)==(i/1000) %monitor acceptance ratio 
    i
    acc/i
  end 
end
acc/I %acceptance ratio

%%% Posterior mean %%%
%%%%%%%%%%%%%%%%%%%%%%
postm=mean(post_draws_cauchy,1);

%%%% Credible intervals %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
diff_mat=(abs(post_draws_cauchy-postm))'; %matrix with I-Ir rows and each row has the L postm_coefs
%L1 credible regions
diff_sq_mean=mean(diff_mat,1);
[sorted, ind]=sort(diff_sq_mean); %the differences in ascending order and their indices
spaghetti_cauchy_ind=ind(1:(0.95*(I-Ir))); %95% of the differences that are the smallest

%%%% Plot results %%%%
%%%%%%%%%%%%%%%%%%%%%%
grayColor = [.7 .7 .7];
figure(3)
lightBlue= [90/255, 106/255, 250/255];
box on
fig.Position(3:4)=[500,400];
hold on
axis manual
axis([0 1 0 1])
for i=1:(0.95*(I-Ir))
  spaghetti_sample=post_draws_cauchy(spaghetti_cauchy_ind(i),:); 
  plot(x,spaghetti_sample,'Color', grayColor) %95% of the smallest differences between posterior draws and posterior mean
end
plot(x,truth, 'LineWidth', 1.75, 'Color', 'black', 'LineStyle','--')
plot(x,postm, 'LineWidth', 1.75, 'Color', lightBlue)
hold off


