% This script implements the whitened pCN algorithm for sampling the posterior
% arising from the OT/HT(alpha) priors based on the Cauchy distribution, in Density Estimation
% with a homogeneously smooth truth. 
% It reproduces the results in Figures D.8 and D.10 in the supplement of Agapiou and Castillo 2024+ 
% "Heavy-tailed Bayesian Nonparametric Adaptation" arXiv:2308.04916


rng(3)
clear all
close all
format long

% Code requires Wavelab850 installed
% Check that computer has Wavelab850 running
run('/Applications/MATLAB_R2022b.app/toolbox/Wavelab850/WavePath.m');


%%%% Data %%%%
%%%%%%%%%%%%%%

K = 2^10; %finest wavelet level 
K1 = 0:(K-1);
K1(1) = 1;
L = 1; % coarsest wavelet level
x=[1/K:1/K:1];


%%%% Wavelet transform %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
QMF8 = MakeONFilter('Daubechies',8); %generate ON wavelet filter for computing wavelet transform
build = @(U) sqrt(K)*IWT_PO(U,L,QMF8); % Perform inverse wavelet transform


%%%% Truth %%%%
%%%%%%%%%%%%%%%
wc_truth=(4*(cos(1:K).^3)).*2.^(-(5/2)*floor(log2(K1))); % define true log-unnormalized-density via
%its coefficients wrt to Daubechies 8 wavelet basis
log_truth=build(wc_truth);
truth=exp(log_truth)';
norm_const=sum(truth)/K;
truth=truth/norm_const; % truth
figure(1)
plot(x,truth, 'LineWidth', 1.75)

%%%% Generate Data, samples from truth %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
n=10^4; %number of observations

% Generate Observations using Rejection Sampling 
data=NaN(1,n);
M=max(truth);
for i=1:n
  y=rand;
  u=rand;
  while (M*u)>truth(ceil(y*K))
    y=rand;
    u=rand;
  end
  data(i)=y;
end
data_disc=ceil(data*K); % discretized data

% Plot frequentist KDE and truth
fig2=figure(2)
fig2.Position(3:4)=[500,400];
axis([0 1 0 4])
box on
hold on
[f,xi] = ksdensity(data, support=[0,1]); 
plot(x,truth,'LineWidth', 1.75, 'Color', 'black', 'LineStyle','--');
plot(xi,f,'LineWidth', 1.75, 'Color', 'red');
xlim=([0,1]);
hold off


%%%% Prior scaling %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
jj = 2.^floor(log2(K1));
% HT(alpha)
%alpha=5;
%scale_vec = ((jj.^(-1/2-alpha)));%prior scaling s_l

% OT
scale_vec=2.^(-(log2(jj)).^(2)); %prior scaling s_l

%%%% Rho or regular posterior %%%%
rho=1; %rho posterior, rho=1 means usual posterior


% Mapping from Gaussian to Cauchy (inverse CDF), needed for whitened
% version of pCN, see Chen et al https://arxiv.org/abs/1803.03344
T = @(xi) K*exp(build(scale_vec.*tan((pi/2)*(1-2*normcdf(xi)))))/sum(exp(build(scale_vec.*tan((pi/2)*(1-2*normcdf(xi)))))); %Cauchy
%white noise representation
% For Gaussian prior use
% T = @(xi) K*exp(build((scale_vec.*xi)'))/sum(exp(build((scale_vec.*xi)')));

%%%% wpCN MCMC parameters %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%
I=200000; %number of iterations

step_size=0.013; %0.0013 for n=10^6, tuned to give around 30% acceptance probability
step_compl=sqrt(1-step_size^2);

post_draws_cauchy=NaN(I,K);

%%%% Initialization %%%%
%%%%%%%%%%%%%%%%%%%%%%%%
old=randn(1,K); %white noise
t_old=T(old); %initialize from the prior
post_draws_cauchy(1,:)=t_old;

acc=0; %acceptance counter

%%%% MCMC %%%%
%%%%%%%%%%%%%%
for i=2:I
  eta=randn(1,K);
  proposal=step_compl*old+step_size*eta;
  t_proposal=T(proposal);
  
  laccept_prob=rho*(sum(log(t_proposal(data_disc)))-sum(log(t_old(data_disc)))); %log of acceptance probability for wpCN

  if log(rand) <= laccept_prob
    old=proposal;
    t_old=t_proposal;
    acc=acc+1;
  end
  post_draws_cauchy(i,:)=t_old;
  if round(i/1000)==(i/1000) %monitor acceptance ratio 
    i
    acc/i
  end
end

acc/I %acceptance ratio


%%%% Posterior mean %%%%
%%%%%%%%%%%%%%%%%%%%%%%%
% Use final 12.5% of iterations only
postm=mean(post_draws_cauchy((7*I/8+1):I,:),1);

%%%% Credible intervals %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
diff_mat=(abs(post_draws_cauchy((7*I/8+1):I,:)-postm))'; %matrix with I/8 rows and each row has the L postm_coefs
%L1 credible set for density estimation!
diff_sq_mean=mean(diff_mat,1);
[sorted, ind]=sort(diff_sq_mean); %the differences in ascending order and their indices
spaghetti_cauchy_ind=ind(1:(0.95*I/8)); %95% of the differences that are the smallest

%%%% Plot results %%%%
%%%%%%%%%%%%%%%%%%%%%%
grayColor = [.7 .7 .7];
lightBlue= [90/255, 106/255, 250/255];
fig=figure(3)
box on
fig.Position(3:4)=[500,400];
hold on
axis manual
axis([0 1 0 4])
for i=1:(0.95*I/8)
  spaghetti_sample=post_draws_cauchy((7*I/8)+spaghetti_cauchy_ind(i),:); 
  plot(x,spaghetti_sample,'Color', grayColor) %95% of the smallest differences between posterior draws and posterior mean
end
plot(x,truth, 'LineWidth', 1.75, 'Color', 'black', 'LineStyle','--')
plot(x,postm, 'LineWidth', 1.75, 'Color', lightBlue)
hold off

