# This script implements sampling via Stan of the posterior
# arising from the OT/HT(alpha) priors based on the Student distribution, 
# in a simple linear inverse problem in white noise, 
# with a homogeneously smooth truth.
# It reproduces the following numerical results of Agapiou and Castillo 2024+ 
# arXiv:2308.04916 "Heavy-tailed Bayesian Nonparametric Adaptation":
# i) 2nd and 3rd columns of Figure 2 in the main article;
# ii) Figures E.1 and E.2 in the supplement.
# To run, it requires the files student-HT.stan and student-OT.stan

library(cmdstanr)
library(posterior)

options(mc.cores = 8)

#############################
#### DEFINITION OF TRUTH ####
N=200 #truncation level

truth_coeffs=((1:N)^(-3/2))*sin(1:N) #coefficients of truth
x=seq(0,1, length.out=N) # grid for spatial domain
t=(1:N-1/2)*pi
aux=outer(x,t)
naive_ift=sqrt(2)*cos(aux)
#we synthesize the coefficients to the spatial domain manually (for plotting)
truth=naive_ift%*%truth_coeffs

#########################################
#### DEFINITION OF NOISY OBSERVATION ####
l=1; # inverse problem l=1 or no inverse problem l=0
n=10^7 # noise precision level

# coefficients of data
y_coeffs=((pi*(1:N-1/2))^(-l))*truth_coeffs+rnorm(N)/sqrt(n)


# Student priors with 3 degrees of freedom
nu=3;
rho=1; # 0<rho<1 for tempered posteriors


#################################
#### HT(alpha) Student prior ####
# Prior variances decay as j^(-1-2alpha)
alpha=5;

# We use Stan to sample each coefficient separately. 
# To this end we have written a Stan file implementing the prior
# and likelihood. Stan returns posterior samples (for each coefficient)
model <- cmdstan_model(stan_file = 'student-HT.stan')
K=1000 #number of iterations in warmup and actual sampling, for each of 
#4 chains that Stan uses by default.

post1_draws_coeffs=matrix(NA,4*K,N) #for storing coefficient posterior samples
post1_draws=matrix(NA, 4*K, N) #for storing samples in spatial domain

# Posterior sampling of each coefficient using Stan
for(j in 1:N)
{
  data=list(y=y_coeffs[j], j=j, sn=sqrt(n), alpha=alpha, nu=nu,l=l, rho=rho)
  fit <- model$sample(data = data, refresh=1000, iter_warmup=K, iter_sampling=K)
  draws<- as_draws_df(fit$draws())
  post1_draws_coeffs[,j]=as.vector(as.data.frame(draws[,2])$f)
}

# Synthesize draws to spatial domain
for(i in 1:(4*K))
{
  post1_draws[i,]=naive_ift%*%post1_draws_coeffs[i,]
}

# Compute posterior mean
pm1=apply(post1_draws, 2, mean)


# Compute 95% (spaghetti) L2-credible bands 
pm1_mat=matrix(data=pm1,nrow=4*K,N, byrow=T)
aux_post1_draws=(post1_draws-pm1_mat)^2
l2distances1=apply(aux_post1_draws, 1, mean)
l2sorted1=sort(l2distances1, index.return=T)
cloud_ind1=l2sorted1$ix[1:(4*K*0.95)]

# results plotted at the end of the script


############################
#### OT - Student prior ####
# Prior variances decay as exp(-a(log(j))^(1+delta))
a=1
delta=0.5

# We use Stan to sample each coefficient separately. 
# To this end we have written a Stan file implementing the prior
# and likelihood. Stan returns posterior samples (for each coefficient)
model <- cmdstan_model(stan_file = 'student-OT.stan')
K=1000 #number of iterations in warmup and actual sampling, for each of 
#4 chains that Stan uses by default.

post2_draws_coeffs=matrix(NA,4*K,N) #for storing coefficient posterior samples
post2_draws=matrix(NA, 4*K, N) #for storing samples in spatial domain

# Posterior sampling using Stan
for(j in 1:N)
{
  data=list(y=y_coeffs[j], j=j, sn=sqrt(n), a=a, delta=delta, nu=3,l=l, rho=rho)
  fit <- model$sample(data = data, refresh=1000, iter_warmup=K, iter_sampling=K)
  draws<- as_draws_df(fit$draws())
  post2_draws_coeffs[,j]=as.vector(as.data.frame(draws[,2])$f)
}

# Synthesize draws to spatial domain
for(i in 1:(4*K))
{
  post2_draws[i,]=naive_ift%*%post2_draws_coeffs[i,]
}

# Compute posterior mean
pm2=apply(post2_draws, 2, mean)


# Compute 95% (spaghetti) L2-credible bands 
pm2_mat=matrix(data=pm2,nrow=4*K,N, byrow=T)
aux_post2_draws=(post2_draws-pm2_mat)^2
l2distances2=apply(aux_post2_draws, 1, mean)
l2sorted2=sort(l2distances2, index.return=T)
cloud_ind2=l2sorted2$ix[1:(4*K*0.95)]


#################################
#### Plot the two posteriors ####
par(mfrow=c(2,1))

#HT(alpha)
plot(x,truth,type="l",  ylim=c(-.25, 2), main=expression("HT, alpha=5"), xlab="", ylab="")
for(li in 1:(4*K*0.95))
{
  lines(x,post1_draws[cloud_ind1[li],], col="lightgray")
}
lines(x,truth,type="l", ylim=c(-.25, 2))
lines(x,pm1, col="4")

#OT
plot(x,truth,type="l",  ylim=c(-.25, 2), main=expression("OT, delta=0.5, a=1"), xlab="", ylab="")
for(li in 1:(4*K*0.95))
{
  lines(x,post2_draws[cloud_ind2[li],], col="lightgray")
}
lines(x,truth,type="l", ylim=c(-.25, 2))
lines(x,pm2, col="4")




