% This script implements the whitened MALA algorithm for sampling the posterior
% arising from the OT prior based on the Cauchy distribution, in the
% white noise model with various spatially inhomogeneous truths. 
% It reproduces the results in Fig D.6 in the supplement of Agapiou and Castillo 2024+ 
% "Heavy-tailed Bayesian Nonparametric Adaptation" arXiv:2308.04916

rng(2)
clear all
close all
format long

% Code requires Wavelab850 installed
% Check that computer has Wavelab850 running
run('/Applications/MATLAB_R2022b.app/toolbox/Wavelab850/WavePath.m');


%%%% Data %%%%
%%%%%%%%%%%%%%%%%%%%%%%%
data=load('Doppler_RegData_WNM'); 
truth=data.truth1'; % true function
wc_y=data.wc_y1'; % wavelet coefficients of observed data, Symmlet 8

K = length(truth); % finest wavelet level
K1 = 0:K-1;
K1(1) = 1;
L = 5; % coarsest wavelet level

x=1/K:1/K:1;

%%%% Wavelet transform %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
QMF8 = MakeONFilter('Symmlet',8); %generate ON wavelet filter for computing wavelet transform
build = @(U) IWT_PO(U,L,QMF8); % Perform inverse wavelet transform


%%%% OT Prior scaling %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
jj = 2.^floor(log2(K1)');
scale_vec=2.^(-(log2(jj)).^(3/2)); %prior scaling s_l

n=1; %truth scaled to have signal to noise ratio 7 when n=1



%%%% wMALA MCMC parameters %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%
h=.000017; % in (0,4]
beta=4*sqrt(h)/(4+h); %step size b
beta_compl=sqrt(1-beta^2);

% Mappings from Gaussian to Cauchy (inverse CDF), needed for whitened
% version of MALA, see Chen et al https://arxiv.org/abs/1803.03344
T = @(xi) tan((pi/2)*(1-2*normcdf(xi))).*scale_vec; %Cauchy
Tinv = @(xi) norminv((1-(2/pi*atan(xi./scale_vec)))/2); %inverse for warm initialization
Lapr = @(xi) -pi*normpdf(xi).*(sec((pi/2)*(1-2*normcdf(xi))).^2); 

Psi = @(xi) (n/2)*(sum(((T(xi)).^2))-2*sum((T(xi).*wc_y)));
DPsi = @(xi) n*Lapr(xi).*(T(xi)-wc_y).*scale_vec;
J = @(xi, xip) Psi(xi)+(h/8)*sum((DPsi(xi).^2))+(sqrt(h)/2)*sum((DPsi(xi).*((xip-beta_compl*xi)/beta)));


I=400000; %number of iterations
Ir=I*(1-1/16); %number of iterations discarded as burn in

post_draws_cauchy=NaN(I-Ir,K);

%%%% MCMC Initialization %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
old=Tinv(wc_y); %warm start seems important for Cauchy!
% for random initialization replace by
%old=randn(1,K)'; %white noise
t_old=T(old);
post_draws_cauchy(1,:)=t_old;

acc=0; %acceptance counter

%%%% MCMC %%%%
%%%%%%%%%%%%%%
for i=2:I
  eta=randn(1,K)';
  
  proposal=beta_compl*old+beta*(eta-sqrt(h)*DPsi(old)/2);
  t_proposal=T(proposal);
  
  laccept_prob=J(old,proposal)-J(proposal,old);

  if log(rand) <= laccept_prob
    old=proposal;
    t_old=t_proposal;
    acc=acc+1;
  end
  if(i>Ir) % only store samples after burnin
    post_draws_cauchy(i-Ir,:)=t_old;
  end
  % display sample every 1000 steps
  if round(i/1000)==(i/1000)  
    i
    acc/i
    figure(1)
    plot(x, build(t_old));colormap('gray')
    view(2)
    axis square
  end
  
end
acc/I % acceptance ratio, tuned to around 30%


%%% Posterior mean %%%
%%%%%%%%%%%%%%%%%%%%%%
pmean=mean(post_draws_cauchy)'; % wavelet coefficients of posterior mean
sqrt(sum((truth-build(pmean)).^2)/K) %L2 norm between truth and posterior mean

%%% Credible regions - Spaghetti plots %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
built_post=zeros(I-Ir, K);
postm_mat=NaN(I-Ir, K);

for i=1:(I-Ir)
    built_post(i,:)=build(post_draws_cauchy(i,:));
    postm_mat(i,:)=pmean;
end

%Generate the differences between posterior draws and posterior mean
diff_coefs_mat=(post_draws_cauchy-postm_mat).^2; %differences between posterior draws and posterior mean
l2dist=mean(diff_coefs_mat'); %calculate the mean in each row of the differences 
[l2sorted, ix]=sort(l2dist); %differences in ascending order and their indices
spaghetti_ind=ix(1:(0.95*(I-Ir))); %find the 95% of the differences that are the smallest


%%%% Plot Results %%%%
%%%%%%%%%%%%%%%%%%%%%%
grayColor = [.7 .7 .7];
%Generate on the same plot: (1)posterior draws (black line), (2)posterior mean (red line), (3)spaghetti plot of 95% of the smallest differences between posterior draws and posterior mean
figure(2)
hold on
for i=1:(0.95*(I-Ir))
  plot(x,built_post(spaghetti_ind(i),:), 'Color', grayColor)
end
plot(x, build(pmean), 'Color', 'b')
plot(x, truth, 'Color', 'k')

hold off

%save('doppler_built_cauchy.mat', 'built_post') %save results

